/**
 * Dryer
 */
#include "main.h"
#include "i2c.h"
#include "sensor.h"
#include "rtos/rtos.h"
#include "ssd1306xled/font6x8.h"
#include "ssd1306xled/font8x16.h"
#include "ssd1306xled/ssd1306xled.h"
#include "ssd1306xled/ssd1306xledtx.h"
#include "tinyavrlib/num2str.h"
#include "bresenham.h"
#include "string.h"

/* Defines */
#define HEATER_PERIOD_MS  1000
#define HEATER_STEP_MS    10
#define HEATER_POWER_MAX  (uint8_t)((HEATER_PERIOD_MS / HEATER_STEP_MS) - 1)

/* Variables */
static volatile struct {
  uint8_t newTempSet: 1;
  uint8_t AHT10state: 1;
  uint8_t ClockRun:   1;
  uint8_t rezerv:     5;
} Flag;
static uint8_t TemperatureSetpoint = 0, oldTempSet = 0;
static aht20_t Sensor;
static bresenham_t Heater;
static clock_t Clock = {0};

/* Function prototypes */
static void board_Init(void);
static void checkTemperatureSetpoint(void);
static void sensorStart(void);
static void sensorGetData(void);
static void LedOn(void);
static void LedOff(void);
static void HeaterCtrl(void);

int main(void) {
  /* Clear flags */
  Flag.newTempSet = 1;
  Flag.AHT10state = 0;
  Flag.ClockRun = 0;

  /* Init all */
  board_Init();
  RTOS_Init();
  I2C_Init();
  AHT20_Init();
  tdelay_ms(40);
  ssd1306_init();

  /* Bresenham init.
   * step - 10 ms, period - 1 sek, heater - off */
  bresenham_Init(&Heater, HEATER_POWER_MAX);
  bresenham_setValue(&Heater, TemperatureSetpoint);
  RTOS_SetTask(HeaterCtrl, HEATER_PERIOD_MS, HEATER_STEP_MS);

  /* Set tasks */
  RTOS_SetTask(sensorStart, 502, 1000);
  RTOS_SetTask(sensorGetData, 952, 1000);
  RTOS_SetTask(checkTemperatureSetpoint, 5, 50);

  /* HeartBeat */
  RTOS_SetTask(LedOn, 1, 1000);
  RTOS_SetTask(LedOff, 201, 1000);

  ssd1306_clear();
//  ssd1306tx_init(ssd1306xled_font6x8data, 0);
  ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 0, 0, "Hello, World! :)");
//  ssd1306_setpos(0, 3);
//  ssd1306tx_string("SSD1306xLED Library");
  tdelay_ms(1000);

  ssd1306_clear();
  ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 8, 0, "Pwr:"); // Set
  ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 8, 2, "  T: xx.x C");
  ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 8, 4, "  H: yy.y %");
  ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 6, ":  :");
  char buffer[9] = {0};
  uint8_t i = 0, d = 0;

  /* Infinity loop */
  do {
    /* Action for new temperature setting */
    if (Flag.newTempSet != 0) {
      Flag.newTempSet = 0;
      if (TemperatureSetpoint != 0) {
        if (oldTempSet == 0) {
        Flag.ClockRun = 1;
          Clock.SS = 0;
          Clock.MM = 0;
          Clock.HH = 0;
        }
        usint2decascii(TemperatureSetpoint, buffer);
        strncpy(buffer+5, " % \0", 4); // C
        ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 0, buffer+3);
      } else {
        Flag.ClockRun = 0;
        ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 0, " Off");
      }
      // set new Heater Power
      bresenham_setValue(&Heater, TemperatureSetpoint);
    }

    /* Output Sensor Data */
    if (Flag.AHT10state == 0) {
      strncpy(buffer, "\0", 6);
      i = Sensor.Temperature / 10;
      d = Sensor.Temperature % 10;
      usint2decascii(i, buffer);
      //strncpy(buffer+5, " C\0", 3);
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 2, buffer+3);
      usint2decascii(d, buffer);
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 72, 2, buffer+4);

      strncpy(buffer, "\0", 6);
      i = Sensor.Humidity / 10;
      d = Sensor.Humidity % 10;
      usint2decascii(i, buffer);
      //strncpy(buffer+5, " %\0", 3);
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 4, buffer+3);
      usint2decascii(d, buffer);
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 72, 4, buffer+4);
    } else {
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 2, "Er");
      ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 48, 4, "Er");
    }

    /* Output Clock */
    strncpy(buffer, "\0", 6);
    usint2decascii(Clock.HH, buffer);
    ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 32, 6, buffer+3);
    usint2decascii(Clock.MM, buffer);
    ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 56, 6, buffer+3);
    usint2decascii(Clock.SS, buffer);
    ssd1306tx_stringxy((uint8_t const *)ssd1306xled_font8x16data, 80, 6, buffer+3);

    RTOS_DispatchTask();

    // nothing to do - sleep, wait for interrupt
#if defined(__GNUC__)
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_mode();
#elif defined(__ICCAVR__)
    MCUCR = 1<<SE;
    __sleep();
#endif
  } while (1);
}

/**
 *  S u b r o u t i n e s
 */

/**
 * @brief Initialize periphery
 */
static void board_Init(void) {
  /* power off Analog Comparator */
  ACSR = ACD;

  /* GPIO */
  DDRB = 0x22; // PB1 - Triac control, PB5 - Green LED
  DDRC = 0x30; // I2C output
  PORTB = 0x03; // close Triac, enable pull-up for D8
  PORTC = 0x3e; // I2C = 1
  PORTD = 0xff; // enable pull-up for D0-D7 (Switch)

  /* Timer0 - RTOS & tdelay_ms() */
  TCCR0B = ((0<<CS02)|(1<<CS01)|(1<<CS00));
  TCNT0 = TIMER0_CNT;
  TIMSK0 |= (1<<TOIE0);

  /* Timer1 - Clock */
  TCCR1B = (1<<CS12 | 0<<CS11 | 1<<CS10);
  TCNT1 = TIMER1_CNT;
  TIMSK1 |= (1<<TOIE1);

  /* Timer2 - refresh screen values */
  //TCCR2 = 0x00;
  //TCNT2 = 0x00;
  //TIMSK |= _BV(TOIE2);

  /* Enable Interrupts */
  ENABLE_INTERRUPT;
}

/**
 * Setup Temperature Set Point by valcoder value
 */
static void checkTemperatureSetpoint(void) {
  static uint8_t prev = 0;
  uint8_t pvalue;

  pvalue = ~(PIND);
  if (pvalue != 0) {
    switch (pvalue) {
    case 1<<PD0:
      TemperatureSetpoint = 20; //55;
      break;
    case 1<<PD1:
      TemperatureSetpoint = 10; //50;
      break;
    case 1<<PD2:
      TemperatureSetpoint = 30; //60;
      break;
    case 1<<PD3:
      TemperatureSetpoint = 40; //65;
      break;
    case 1<<PD4:
      TemperatureSetpoint = 50; //70;
      break;
    case 1<<PD5:
      TemperatureSetpoint = 60; //75;
      break;
    case 1<<PD6:
      TemperatureSetpoint = 70; //80;
      break;
    case 1<<PD7:
      TemperatureSetpoint = 80; //85;
      break;
    default:
      TemperatureSetpoint = 0;
    }
  } else {
    pvalue = (~(PINB)) & (1<<PB0);
    if (pvalue != 0) {
      TemperatureSetpoint = 90;
    } else {
      TemperatureSetpoint = 0;
    }
  }

  if (prev != TemperatureSetpoint) {
    Flag.newTempSet = 1;
    oldTempSet = prev;
    prev = TemperatureSetpoint;
  }
}

static void sensorStart(void) {
  aht20_st_t status;

  status = AHT20_StartMeasure();

  if (status != AHT_St_OK) {
    Flag.AHT10state = 1;
  } else {
    Flag.AHT10state = 0;
  }
}

static void sensorGetData(void) {
  aht20_st_t status;

  status = AHT20_GetData(&Sensor);

  if (status != AHT_St_OK) {
    Flag.AHT10state = 1;
  } else {
    Flag.AHT10state = 0;
  }
}

static void LedOn(void) {
  PORTB |= (1<<PB5);
}

static void LedOff(void) {
  PORTB &= ~(1<<PB5);
}

/**
 * On/Off heater by Bresenham algorithm
 */
static void HeaterCtrl(void) {
  if (bresenham_getNext(&Heater)) {
    PORTB &= ~(1<<PB1); // Heater ON
  } else {
    PORTB |= (1<<PB1);// Heater OFF
  }
}

/**
 * I n t e r r u p t s
 */
#if defined(__ICCAVR__)
#pragma vector=TIMER1_OVF_vect
__interrupt void TIMER1_OVF_ISR(void)
#elif defined(__GNUC__)
ISR(TIMER1_OVF_vect)
#endif
{
  /* reload timer counter */
  TCNT1 = TIMER1_CNT;

  /* Clock */
  if (Flag.ClockRun != 0) {
    Clock.SS ++;
    if (Clock.SS > 59) {
      Clock.SS = 0;
      Clock.MM ++;
      if (Clock.MM > 59) {
        Clock.MM = 0;
        Clock.HH ++;
      }
    }
  }
}

#if defined(__GNUC__)
/**
 * @brief заглушка для неиспользуемых прерываний
 */
ISR(__vector_default,ISR_NAKED) {
  reti();
}
#endif
